/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.bus;

import java.io.IOException;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.Protocol;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush;
import org.objectweb.dream.protocol.messagePassing.MessagePassingProtocol;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * Implementation of the Bus protocol.
 */
public class BusProtocolImpl extends AbstractComponent implements BusProtocol
{

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected MessagePassingProtocol lowerLevelProtocolItf;
  protected MessageManager         messageManagerItf;
  protected LocalAccessManager     localAccessManagerItf;

  // ---------------------------------------------------------------------------
  // Implementation of the BusProtocol interface
  // ---------------------------------------------------------------------------

  /**
   * @see BusProtocol#bind(ExportIdentifier, IncomingPush, Map)
   */
  public BusOutgoingPush bind(ExportIdentifier exportId,
      IncomingPush toClientPush, Map<String, Object> hints)
      throws InvalidExportIdentifierException, BindException
  {
    if (!(exportId instanceof BusExportIdentifier))
    {
      throw new InvalidExportIdentifierException(
          "Must be a BusExportIdentifier");
    }
    BusExportIdentifier busExportIdentifier = (BusExportIdentifier) exportId;
    int localAddress;
    Object o = hints.get(LOCAL_ADDRESS_HINT_KEY);
    if (o != null)
    {
      if (!(o instanceof Number))
      {
        throw new BindException("Invalid local address hint, not a number.");
      }
      localAddress = ((Number) o).intValue();
    }
    else
    {
      localAddress = localAccessManagerItf.findLocalAddress(busExportIdentifier
          .getNextExportIds());
      if (localAddress == -1)
      {
        throw new BindException(
            "Unable to find local process address in given export identifier");
      }
    }
    Session session = new Session(busExportIdentifier, localAddress,
        toClientPush);
    try
    {
      session.lowerOutgoingPush = localAccessManagerItf.exportBus(
          busExportIdentifier.getNextExportIds()[localAddress], session);
    }
    catch (ExportException e)
    {
      throw new BindException("Unable to export access to this protocol",
          exportId, e);
    }
    return session;
  }

  /**
   * @see ChannelProtocol#createExportIdentifier(Map, ExportIdentifier[])
   */
  public ExportIdentifier createExportIdentifier(Map info,
      ExportIdentifier[] next) throws InvalidExportIdentifierException
  {
    return new BusExportIdentifier(next);
  }

  // ---------------------------------------------------------------------------
  // Inner classes
  // ---------------------------------------------------------------------------

  protected class Session implements BusOutgoingPush, IncomingPush
  {
    protected int                        myAddress;
    protected BusExportIdentifier        id;
    protected ExportIdentifier[]         lowerIds;
    protected boolean                    closed             = false;

    protected MessagePassingOutgoingPush lowerOutgoingPush;
    protected IncomingPush               upperIncomingPush;

    protected Session(BusExportIdentifier id, int myAddress,
        IncomingPush upperIncomingPush)
    {
      this.id = id;
      lowerIds = id.ids;
      this.myAddress = myAddress;
      this.upperIncomingPush = upperIncomingPush;
    }

    // -------------------------------------------------------------------------
    // Implementation of the BusOutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see BusOutgoingPush#outgoingPush(Message, int)
     */
    public void outgoingPush(Message message, int to) throws PushException
    {
      if (to > lowerIds.length || to < 0)
      {
        throw new IOPushException("Invalid destination number : " + to);
      }
      synchronized (this)
      {
        if (closed)
        {
          throw new IOPushException("Bus session is closed");
        }
      }
      if (to == myAddress)
      {
        upperIncomingPush.incomingPush(message);
      }
      else
      {
        try
        {
          lowerOutgoingPush.outgoingPush(message, lowerIds[to]);
        }
        catch (InvalidExportIdentifierException e)
        {
          throw new PushException(e);
        }
      }
    }

    /**
     * @see BusOutgoingPush#outgoingClose(IncomingPush)
     */
    public synchronized void outgoingClose(IncomingPush incomingPush)
        throws IOException
    {
      lowerOutgoingPush.outgoingClose(this);
      closed = true;
    }

    /**
     * @see BusOutgoingPush#getLocalAddress()
     */
    public int getLocalAddress()
    {
      return myAddress;
    }

    /**
     * @see BusOutgoingPush#getBusSize()
     */
    public int getBusSize()
    {
      return lowerIds.length;
    }

    // -------------------------------------------------------------------------
    // Implementation of the IncomingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see IncomingPush#incomingPush(Message)
     */
    public void incomingPush(Message message) throws PushException
    {
      // just forward the message.
      upperIncomingPush.incomingPush(message);
    }

    /**
     * @see IncomingPush#incomingClosed(Object, Exception)
     */
    public synchronized void incomingClosed(Object outgoingPush,
        Exception exception)
    {
      // never called on Message Passing Protocol
      Error.bug(logger);
    }

  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME,
        ChannelProtocol.LOWER_PROTOCOL_ITF_NAME, LocalAccessManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(Protocol.LOWER_PROTOCOL_ITF_NAME))
    {
      lowerLevelProtocolItf = (MessagePassingProtocol) serverItf;
    }
    else if (clientItfName.equals(LocalAccessManager.ITF_NAME))
    {
      localAccessManagerItf = (LocalAccessManager) serverItf;
    }
    super.bindFc(clientItfName, serverItf);
  }
}
