/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Romain Lenglet
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.queue;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InterruptedPullException;
import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.message.Message;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * An outgoing handler that {@link #pull() returns} messages removed from a
 * {@link BufferRemove buffer} bound to it. This component uses
 * {@link BufferRemove#remove() blocking calls for removing} messages from the
 * buffer.
 */
public class OutgoingHandlerBlockingImpl extends AbstractComponent
    implements
      Pull
{

  // ------------------------------------------------------------------------
  // Constants.
  // ------------------------------------------------------------------------

  /**
   * The names of client interfaces.
   */
  protected static final String[] CLIENT_INTERFACE_NAMES = {BufferRemove.ITF_NAME};

  // ------------------------------------------------------------------------
  // Client interfaces.
  // ------------------------------------------------------------------------

  /**
   * The buffer from which pulled messages are removed.
   */
  protected BufferRemove          bufferRemove;

  // ------------------------------------------------------------------------
  // Implementation of the Pull interface.
  // ------------------------------------------------------------------------

  /**
   * @see Pull#pull()
   */
  public Message pull() throws PullException
  {
    try
    {
      return bufferRemove.remove();
    }
    catch (InterruptedException e)
    {
      throw new InterruptedPullException(
          "interrupted while removing message from the buffer", e);
    }
  }

  // ------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return CLIENT_INTERFACE_NAMES;
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(BufferRemove.ITF_NAME))
    {
      bufferRemove = (BufferRemove) serverItf;
    }
  }
}
