/**
 * Dream
 * Copyright (C) 2003 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.queue;

import java.util.LinkedList;
import java.util.List;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InterruptedPullException;
import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.Push;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.util.EmptyStringArray;

/**
 * High performance implementation of a Push/Pull queue. This queue has the
 * following properties :
 * <ul>
 * <li>FIFO</li>
 * <li>Thread safe</li>
 * <li>Has no maximum capacity</li>
 * <li>{@link #pull} is blocking if queue is empty.</li>
 * </ul>
 */
public class PushPullQueueFastImpl extends AbstractComponent
    implements
      Push,
      Pull,
      BufferAttributeController
{

  // ------------------------------------------------------------------------
  // Fields.
  // ------------------------------------------------------------------------

  protected List<Message> list = new LinkedList<Message>();

  // ------------------------------------------------------------------------
  // Attributes
  // ------------------------------------------------------------------------

  protected int  size    = 0;

  // ------------------------------------------------------------------------
  // Implementation of the Push interface.
  // ------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public synchronized void push(Message message)
  {
    size++;
    list.add(message);
    if (size == 1)
    {
      notify();
    }
  }

  // ------------------------------------------------------------------------
  // Implementation of the Pull interface.
  // ------------------------------------------------------------------------

  /**
   * @see Pull#pull()
   */
  public synchronized Message pull() throws PullException
  {
    while (size == 0)
    {
      try
      {
        wait();
      }
      catch (InterruptedException e)
      {
        throw new InterruptedPullException(e);
      }
    }
    size--;
    return list.remove(0);
  }

  // ------------------------------------------------------------------------
  // Implementation of the BufferAttributeController interface.
  // ------------------------------------------------------------------------

  /**
   * @see BufferAttributeController#getCurrentSize()
   */
  public int getCurrentSize()
  {
    return size;
  }

  /**
   * @see BufferAttributeController#getMaxCapacity()
   */
  public int getMaxCapacity()
  {
    return 0;
  }

  /**
   * @see BufferAttributeController#setMaxCapacity(int)
   */
  public void setMaxCapacity(int maxCapacity)
  {
    // do nothing no max capacity
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }

}