/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.queue;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import org.objectweb.dream.message.AbstractChunk;

/**
 * Chunk interface for messages that have to be sequentially ordered.
 */
public class SequenceNumberChunk extends AbstractChunk<SequenceNumberChunk>
{
  private static final long  serialVersionUID = 3616730456838386232L;

  /** The default name for chunks of this type. */
  public static final String DEFAULT_NAME     = "sequenceNumberChunk";

  /** The sequence number. */
  private long               sn;

  /**
   * If the given <code>chunkName</code> is equals to the default name returns
   * {@link #DEFAULT_NAME }, returns the given string otherwise.
   * 
   * @param chunkName the chunk name to filter.
   * @return {@link #DEFAULT_NAME }or the given <code>chunkName</code>.
   */
  public static String filterChunkName(String chunkName)
  {
    if (DEFAULT_NAME == chunkName || DEFAULT_NAME.equals(chunkName))
    {
      return DEFAULT_NAME;
    }
    return chunkName;
  }

  /** Default constructor */
  public SequenceNumberChunk()
  {
  }

  /**
   * @param sequenceNumber the sequence number of this chunk.
   */
  public SequenceNumberChunk(long sequenceNumber)
  {
    this.sn = sequenceNumber;
  }

  // ---------------------------------------------------------------------------
  // Chunk methods
  // ---------------------------------------------------------------------------

  /**
   * Returns the sequence number of this chunk.
   * 
   * @return the sequence number of this chunk.
   */
  public final long getSequenceNumber()
  {
    return sn;
  }

  /**
   * Sets the sequence number of this chunk.
   * 
   * @param sn the sequence number to set.
   */
  public final void setSequenceNumber(long sn)
  {
    this.sn = sn;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Chunk interface
  // ---------------------------------------------------------------------------

  /**
   * @see AbstractChunk#transfertStateTo(AbstractChunk)
   */
  public void transfertStateTo(SequenceNumberChunk newInstance)
  {
    newInstance.setSequenceNumber(getSequenceNumber());
  }

  /**
   * @see AbstractChunk#newChunk()
   */
  public SequenceNumberChunk newChunk()
  {
    return new SequenceNumberChunk();
  }

  /**
   * @see org.objectweb.dream.pool.Recyclable#recycle()
   */
  public void recycle()
  {
    sn = 0;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Externalizable method
  // ---------------------------------------------------------------------------

  /**
   * @see java.io.Externalizable#readExternal(ObjectInput)
   */
  public void readExternal(ObjectInput in) throws IOException,
      ClassNotFoundException
  {
    sn = in.readLong();
  }

  /**
   * @see java.io.Externalizable#writeExternal(ObjectOutput)
   */
  public void writeExternal(ObjectOutput out) throws IOException
  {
    out.writeLong(sn);
  }

}
