/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.util.EmptyStringArray;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * Basic implementation of the {@link Condition }interface. This implementation
 * uses a non-renetrant mutex.
 * <p>
 * <b>Note: Inspired by Doug Lea's implementation. </b>
 */
public class ConditionImpl extends AbstractComponent implements Condition

{

  // Client interfaces
  protected Mutex mutexItf;

  // ---------------------------------------------------------------------------
  // Implementation of the Condition interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.synchro.Condition#await()
   */
  public void await() throws InterruptedException
  {
    if (Thread.interrupted())
    {
      throw new InterruptedException();
    }
    try
    {
      synchronized (this)
      {
        mutexItf.unlock();
        try
        {
          wait();
        }
        catch (InterruptedException ex)
        {
          notify();
          throw ex;
        }
      }
    }
    finally
    {
      // Must ignore interrupt on re-acquire
      boolean interrupted = false;
      for (;;)
      {
        try
        {
          mutexItf.lock();
          break;
        }
        catch (InterruptedException ex)
        {
          interrupted = true;
        }
      }
      if (interrupted)
      {
        Thread.currentThread().interrupt();
      }
    }
  }

  /**
   * @see org.objectweb.dream.synchro.Condition#timedwait(long)
   */
  public boolean timedwait(long msecs) throws InterruptedException
  {
    if (Thread.interrupted())
      throw new InterruptedException();
    boolean success = false;
    try
    {
      synchronized (this)
      {
        mutexItf.unlock();
        try
        {
          if (msecs > 0)
          {
            long start = System.currentTimeMillis();
            wait(msecs);
            success = System.currentTimeMillis() - start <= msecs;
          }
        }
        catch (InterruptedException ex)
        {
          notify();
          throw ex;
        }
      }
    }
    finally
    {
      // Must ignore interrupt on re-acquire
      boolean interrupted = false;
      for (;;)
      {
        try
        {
          mutexItf.lock();
          break;
        }
        catch (InterruptedException ex)
        {
          interrupted = true;
        }
      }
      if (interrupted)
      {
        Thread.currentThread().interrupt();
      }
    }
    return success;
  }

  /**
   * @see org.objectweb.dream.synchro.Condition#signal()
   */
  public synchronized void signal()
  {
    notify();
  }

  /**
   * @see org.objectweb.dream.synchro.Condition#signalAll()
   */
  public synchronized void signalAll()
  {
    notifyAll();
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Mutex.ITF_NAME))
    {
      mutexItf = (Mutex) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return EmptyStringArray.EMPTY_STRING_ARRAY;
  }

}